#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

#include "debug.h"
#include "kprintf.h"
#include "memory.h"
#include "utils.h"

#define MAX_FRAMES (HEAP_END / PAGE_SIZE)
#define NB_FRAMES  (mem_size / PAGE_SIZE)

enum {
	FREE,
	USED,
};

static uint8_t frame_table[MAX_FRAMES];

static int32_t find_next_block(size_t nb_frames)
{
	for (uint32_t i = CEIL(HEAP_START, PAGE_SIZE);
	     i + nb_frames < NB_FRAMES; i++) {
		uint32_t j;
		for (j = 0; frame_table[i + j] == FREE && j < nb_frames; j++)
			;
		if (j == nb_frames)
			return i;
		i += j;
	}
	return -1;
}

void *alloc_frames(size_t size)
{
	const uint32_t nb_frames = CEIL(size, PAGE_SIZE);
	const int i = find_next_block(nb_frames);
	if (i < 0) {
		kprintf(KERN_WARNING "Not enough frames available\n");
		return NULL;
	}
	for (size_t j = 0; j < nb_frames; j++) {
		assert(j + i < NB_FRAMES);
		frame_table[j + i] = USED;
	}
	return (void *)(i * PAGE_SIZE);
}

int free_frames(void *frame_ptr, size_t size)
{
	const uint32_t nb_frames = CEIL(size, PAGE_SIZE);
	const uint32_t start = (uint32_t)frame_ptr / PAGE_SIZE;

	if (start > NB_FRAMES) {
		kprintf(KERN_WARNING "Address out of range\n");
		return -1;
	} else if ((uint32_t)frame_ptr % PAGE_SIZE) {
		kprintf(KERN_WARNING "Invalid address\n");
		return -1;
	} else if (start + nb_frames > NB_FRAMES) {
		kprintf(KERN_WARNING "Invalid number of frames\n");
		return -1;
	}
	for (size_t i = start; i < start + nb_frames; i++)
		frame_table[i] = FREE;
	return 0;
}
